from functools import reduce
from operator import mul

import numpy as np


# math.prod in 3.8
# https://docs.python.org/3/library/math.html#math.prod
def _prod(a):
    return reduce(mul, a, 1)


def dot(a, b):
    """Take arrays `a` and `b` and form the dot product between the last axis of `a` and
    the first of `b`.
    """
    b = np.asarray(b)
    return np.dot(a, b.reshape(b.shape[0], -1)).reshape(a.shape[:-1] + b.shape[1:])


def solve(A, x):
    """Solves a linear equation system with a matrix of shape (n, n) and an array of
    shape (n, ...). The output has the same shape as the second argument.
    """
    # https://stackoverflow.com/a/48387507/353337
    x = np.asarray(x)
    return np.linalg.solve(A, x.reshape(x.shape[0], -1)).reshape(x.shape)


def sum_at(a, indices, minlength: int):
    """Sums up values `a` with `indices` into an output array of at least length
    `minlength` while treating dimensionality correctly. It's a lot faster than numpy's
    own np.add.at (see
    https://github.com/numpy/numpy/issues/5922#issuecomment-511477435).

    Typically, `indices` will be a one-dimensional array; `a` can have any
    dimensionality. In this case, the output array will have shape (minlength,
    a.shape[1:]).

    `indices` may have arbitrary shape, too, but then `a` has to start out the same.
    (Those dimensions are flattened out in the computation.)
    """
    a = np.asarray(a)
    indices = np.asarray(indices)

    if len(a.shape) < len(indices.shape):
        raise RuntimeError(
            f"a.shape = {a.shape}, indices.shape = {indices.shape}, "
            "but len(a.shape) >= len(indices.shape) is required."
        )

    m = len(indices.shape)
    assert indices.shape == a.shape[:m]

    out_shape = (minlength, *a.shape[m:])

    indices = indices.reshape(-1)
    a = a.reshape(_prod(a.shape[:m]), _prod(a.shape[m:]))

    return np.array(
        [
            np.bincount(indices, weights=a[:, k], minlength=minlength)
            for k in range(a.shape[1])
        ]
    ).T.reshape(out_shape)


def add_at(a, indices, b):
    a = np.asarray(a)
    indices = np.asarray(indices)
    b = np.asarray(b)

    m = len(indices.shape)
    assert a.shape[1:] == b.shape[m:]
    a += sum_at(b, indices, a.shape[0])


def subtract_at(a, indices, b):
    b = np.asarray(b)
    add_at(a, indices, -b)


def unique_rows(a, return_inverse=False, return_counts=False):
    # The numpy alternative `np.unique(a, axis=0)` is slow; cf.
    # <https://github.com/numpy/numpy/issues/11136>.
    a = np.asarray(a)
    if not np.issubdtype(a.dtype, np.integer):
        raise ValueError(f"Input array must be integer type, got {a.dtype}.")

    a_shape = a.shape
    a = a.reshape(a.shape[0], -1)

    b = np.ascontiguousarray(a).view(np.dtype((np.void, a.dtype.itemsize * a.shape[1])))
    out = np.unique(b, return_inverse=return_inverse, return_counts=return_counts)
    # out[0] are the sorted, unique rows
    if return_inverse or return_counts:
        out = (out[0].view(a.dtype).reshape(-1, *a_shape[1:]), *out[1:])
    else:
        out = out.view(a.dtype).reshape(-1, *a_shape[1:])

    return out



from collections import namedtuple
from typing import Optional

import numpy as np
import scipy
import scipy.sparse.linalg

Info = namedtuple("KrylovInfo", ["success", "xk", "resnorms", "errnorms"])


def cg(
    A,
    b,
    x0=None,
    tol: float = 1e-05,
    maxiter: Optional[int] = None,
    M=None,
    callback=None,
    atol: Optional[float] = 0.0,
    exact_solution=None,
):
    resnorms = []

    if exact_solution is None:
        errnorms = None
    else:
        err = exact_solution - x0
        errnorms = [np.sqrt(np.dot(err, err))]

    def cb(xk):
        if callback is not None:
            callback(xk)

        r = b - A @ xk
        Mr = r if M is None else M @ r
        resnorms.append(np.sqrt(np.dot(r, Mr)))

        if exact_solution is not None:
            err = exact_solution - x0
            errnorms.append(np.sqrt(np.dot(err, err)))

    x, info = scipy.sparse.linalg.cg(
        A, b, x0=x0, tol=tol, maxiter=maxiter, M=M, atol=atol, callback=cb
    )

    success = info == 0

    resnorms = np.array(resnorms)
    if errnorms is not None:
        errnorms = np.array(errnorms)

    return x if success else None, Info(success, x, resnorms, errnorms)


def gmres(
    A,
    b,
    x0=None,
    tol: float = 1e-05,
    restart: Optional[int] = None,
    maxiter: Optional[int] = None,
    M=None,
    callback=None,
    atol: Optional[float] = 0.0,
    exact_solution=None,
):
    resnorms = []

    if exact_solution is None:
        errnorms = None
    else:
        err = exact_solution - x0
        errnorms = [np.sqrt(np.dot(err, err))]

    def cb(xk):
        if callback is not None:
            callback(xk)

        r = b - A @ xk
        Mr = r if M is None else M @ r
        resnorms.append(np.sqrt(np.dot(r, Mr)))

        if exact_solution is not None:
            err = exact_solution - x0
            errnorms.append(np.sqrt(np.dot(err, err)))

    x, info = scipy.sparse.linalg.gmres(
        A,
        b,
        x0=x0,
        tol=tol,
        restart=restart,
        maxiter=maxiter,
        M=M,
        atol=atol,
        callback=cb,
        callback_type="x",
    )

    success = info == 0

    resnorms = np.array(resnorms)
    if errnorms is not None:
        errnorms = np.array(errnorms)

    return x if success else None, Info(success, x, resnorms, errnorms)


def minres(
    A,
    b,
    x0=None,
    shift: float = 0.0,
    tol: float = 1e-05,
    maxiter: Optional[int] = None,
    M=None,
    callback=None,
    exact_solution=None,
):
    resnorms = []

    if exact_solution is None:
        errnorms = None
    else:
        err = exact_solution - x0
        errnorms = [np.sqrt(np.dot(err, err))]

    def cb(xk):
        if callback is not None:
            callback(xk)

        r = b - A @ xk
        Mr = r if M is None else M @ r
        resnorms.append(np.sqrt(np.dot(r, Mr)))

        if exact_solution is not None:
            err = exact_solution - x0
            errnorms.append(np.sqrt(np.dot(err, err)))

    x, info = scipy.sparse.linalg.minres(
        A, b, x0=x0, shift=shift, tol=tol, maxiter=maxiter, M=M, callback=cb
    )

    success = info == 0

    resnorms = np.array(resnorms)
    if errnorms is not None:
        errnorms = np.array(errnorms)

    return x if success else None, Info(success, x, resnorms, errnorms)



import numpy as np
import pytest

import npx

np.random.seed(0)


def test_dot():
    a = np.random.rand(1, 2, 3)
    b = np.random.rand(3, 4, 5)
    c = npx.dot(a, b)
    assert c.shape == (1, 2, 4, 5)


def test_solve():
    a = np.random.rand(3, 3)
    b = np.random.rand(3, 4, 5)
    c = npx.solve(a, b)
    assert c.shape == b.shape


def test_sum_at():
    a = [1.0, 2.0, 3.0]
    idx = [0, 1, 0]
    out = npx.sum_at(a, idx, minlength=4)

    tol = 1.0e-13
    ref = np.array([4.0, 2.0, 0.0, 0.0])
    assert np.all(np.abs(out - ref) < (1 + np.abs(ref)) * tol)


def test_add_at():
    a = [1.0, 2.0, 3.0]
    idx = [0, 1, 0]
    out = np.zeros(2)
    npx.add_at(out, idx, a)

    tol = 1.0e-13
    ref = np.array([4.0, 2.0])
    assert np.all(np.abs(out - ref) < (1 + np.abs(ref)) * tol)


def test_subtract_at():
    a = [1.0, 2.0, 3.0]
    idx = [0, 1, 0]
    out = np.ones(2)
    npx.subtract_at(out, idx, a)

    tol = 1.0e-13
    ref = np.array([-3.0, -1.0])
    assert np.all(np.abs(out - ref) < (1 + np.abs(ref)) * tol)


def test_unique_rows():
    a = [1, 2, 1]
    a_unique = npx.unique_rows(a)
    assert np.all(a_unique == [1, 2])

    a = [[1, 2], [1, 4], [1, 2]]
    a_unique, inv, count = npx.unique_rows(a, return_inverse=True, return_counts=True)
    assert np.all(a_unique == [[1, 2], [1, 4]])
    assert np.all(inv == [0, 1, 0])
    assert np.all(count == [2, 1])

    a_unique = npx.unique_rows(a)
    assert np.all(a_unique == [[1, 2], [1, 4]])

    # entries are matrices
    # fails for some reason. keep an eye on
    # <https://stackoverflow.com/q/67128631/353337>
    a = [[[3, 4], [-1, 2]], [[3, 4], [-1, 2]]]
    a_unique = npx.unique_rows(a)
    assert np.all(a_unique == [[[3, 4], [-1, 2]]])

    a = [1.1, 2.2]
    with pytest.raises(ValueError):
        a_unique = npx.unique_rows(a)
